/*
 * include/asm-microblaze/processor.h
 *
 *  Copyright (C) 2003       John Williams <jwilliams@itee.uq.edu.au>
 *  Copyright (C) 2001,2002  NEC Corporation
 *  Copyright (C) 2001,2002  Miles Bader <miles@gnu.org>
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License.  See the file COPYING in the main directory of this
 * archive for more details.
 *
 * Written by Miles Bader <miles@gnu.org>
 * Microblaze port by John Williams
 */

#ifndef __MICROBLAZE_PROCESSOR_H__
#define __MICROBLAZE_PROCESSOR_H__

#include <linux/config.h>

#include <asm/ptrace.h>
#include <asm/current.h>
#include <asm/entry.h>

/* Some code expects `segment' stuff to be defined here.  */
#include <asm/segment.h>

/* Machine State Register (MSR) Fields */
#define MSR_BE  (1<<0)
#define MSR_IE  (1<<1)
#define MSR_C   (1<<2)
#define MSR_BIP (1<<3)
#define MSR_FSL (1<<4)
#define MSR_ICE (1<<5)
#define MSR_DZ  (1<<6)
#define MSR_DCE (1<<7)
#define MSR_EE  (1<<8)
#define MSR_EIP (1<<9)
#define MSR_UM (1<<11)		/* User Mode */
#define MSR_UMS (1<<12)		/* User Mode Save */
#define MSR_VM  (1<<13)		/* Virtual Mode */
#define MSR_VMS (1<<14)		/* Virtual Mode Save */
#define MSR_CC   (1<<31)

#define MSR_		MSR_EE

#define MSR_KERNEL      MSR_|MSR_VM
#define MSR_USER	MSR_KERNEL|MSR_UM|MSR_IE
#define MSR_KERNEL_VMS  MSR_|MSR_VMS
#define MSR_USER_VMS	MSR_KERNEL_VMS|MSR_UMS|MSR_IE

/* Exception State Register (ESR) Fields */
#define	  ESR_DIZ	(1<<11)		/* Zone Protection */
#define	  ESR_S		(1<<10)		/* Store instruction */

/* Floating Point Status Register (FSR) Bits */
#define FSR_IO		(1<<4)		/* Invalid operation */
#define FSR_DZ		(1<<3)		/* Divide-by-zero */
#define FSR_OF		(1<<2)		/* Overflow */
#define FSR_UF		(1<<1)		/* Underflow */
#define FSR_DO		(1<<0)		/* Denormalized operand error */


#ifdef NO_MM
/*
 * The only places this is used seem to be horrible bletcherous kludges,
 * so we just define it to be as large as possible.
 */
#define TASK_SIZE	(0xFFFFFFFF)
#else
/*
 * This is used to define STACK_TOP, and with MMU it must be below
 * kernel base to select the correct PGD when handling MMU exceptions.
 */
#define TASK_SIZE	(CONFIG_KERNEL_START)
#endif

/* This decides where the kernel will search for a free chunk of vm
 * space during mmap's.
 */
#define TASK_UNMAPPED_BASE	(TASK_SIZE / 8 * 3)


//#define THREAD_SIZE	(2*PAGE_SIZE)


#ifndef __ASSEMBLY__

/*
 * Default implementation of macro that returns current
 * instruction pointer ("program counter").
 */
#define current_text_addr()	({ __label__ _l; _l: &&_l;})


/*
 * Bus types
 */
#define EISA_bus 0
#define EISA_bus__is_a_macro /* for versions in ksyms.c */
#define MCA_bus 0
#define MCA_bus__is_a_macro /* for versions in ksyms.c */

/* If you change this, you must change the associated assembly-languages
   constants defined below, THREAD_*.  */
struct thread_struct {
	/* kernel stack pointer (must be first field in structure) */
	unsigned long  ksp;
#ifndef NO_MM
	mm_segment_t	fs;		/* for get_fs() validation */
	void		*pgdir;		/* root of page-table tree */
	struct pt_regs	*regs;		/* Pointer to saved register state */
#endif
};

#ifdef NO_MM
#define INIT_THREAD { sizeof init_stack + (unsigned long)init_stack }
#else
#define INIT_THREAD { \
	.ksp   = sizeof init_stack + (unsigned long)init_stack, \
	.fs    = KERNEL_DS, \
	.pgdir = swapper_pg_dir, \
}
#endif


/* Do necessary setup to start up a newly executed thread.  */
#ifdef NO_MM
extern inline void start_thread (struct pt_regs *regs,
				 unsigned long pc, unsigned long usp)
{
	/* Thread will be started as a return from syscall
		rtbd r14, 4.
	   Therefore, we adjust start PC address by negative 4 */
	regs->pc = pc-4;
	regs->sp = usp;
	regs->kernel_mode = 0;
}
#else
void start_thread (struct pt_regs *regs,
		   unsigned long pc, unsigned long usp);
#endif /* NO_MM */

/* Free all resources held by a thread. */
extern inline void release_thread (struct task_struct *dead_task)
{
}

extern int kernel_thread(int (*fn)(void *), void * arg, unsigned long flags);

/* Free current thread data structures etc.  */
static inline void exit_thread (void)
{
}

/* Return saved (kernel) PC of a blocked thread.  */
#define thread_saved_pc(tsk)	\
	((tsk)->thread.regs? (tsk)->thread.regs->r15: 0)

unsigned long get_wchan (struct task_struct *p);


/* Return some info about the user process TASK.  */
#define task_tos(task)	((unsigned long)(task) + KERNEL_STACK_SIZE)
#define task_regs(task) ((struct pt_regs *)task_tos (task) - 1)

#define task_pt_regs(tsk) (((struct pt_regs *)(THREAD_SIZE + task_stack_page(tsk))) - 1)
#define task_pt_regs_plus_args(tsk) (((void *)task_pt_regs(tsk)) - STATE_SAVE_ARG_SPACE)

#define task_thread_info(task) (task)->thread_info
#define task_stack_page(task) ((void*)((task)->thread_info))
#define task_sp(task)	(task_regs (task)->sp)
#define task_pc(task)	(task_regs (task)->pc)
/* Grotty old names for some.  */
#define KSTK_EIP(task)	task_pc (task)
#define KSTK_ESP(task)	task_sp (task)

#define cpu_relax() barrier()

/* Allocation and freeing of basic task resources. */
/*#define alloc_task_struct() \
   ((struct task_struct *)__get_free_pages(GFP_KERNEL,1))
#define free_task_struct(p)	free_pages ((unsigned long)(p),1)
#define get_task_struct(tsk)	atomic_inc (&mem_map[MAP_NR(tsk)].count)
*/

//#define init_task	(init_task_union.task)
//#define init_stack	(init_task_union.stack)

/* FIXME */
#define prepare_to_copy(tsk)	do {} while(0)
#define deactivate_mm(tsk,mm)	do { } while (0)


#else /* !__ASSEMBLY__ */


#define THREAD_KSP	0


#endif /* __ASSEMBLY__ */


#endif /* __MICROBLAZE_PROCESSOR_H__ */
